==> platform-3.0.10.jar: hara/platform/exchange/base.clj <==
(ns hara.platform.exchange.base
  (:require [hara.protocol.exchange :as protocol.exchange]
            [hara.protocol.component :as protocol.component]
            [hara.core.component :as component]
            [hara.data.table :as table]
            [hara.platform.exchange.common :as common]))

(defn match-pattern
  "creates a re-pattern for the rabbitmq regex string
 
   (match-pattern \"*\" \"hello\")
   => true
 
   (match-pattern \".*.\" \".hello.\")
   => true"
  {:added "3.0"}
  [pattern input]
  (-> pattern
      (.replaceAll "\\." "\\\\\\.")
      (.replaceAll "\\*" ".*")
      (re-pattern)
      (re-find input)
      (= input)))

(defn route?
  "checks if a message will be routed
 
   (route? {:type \"fanout\"} {} {})
   => true
 
   (route? {:type \"topic\"}
           {:key \"user.account.login\"}
           {:routing-key \"*.account.*\"})
   => true
 
   (route? {:type \"header\"}
           {:headers {\"account\" \"login\"}}
           {:arguments {\"account\" \"*\"}})
   => true"
  {:added "3.0"}
  [{:keys [type] :as meta}
   {:keys [key headers] :as package}
   {:keys [routing-key arguments] :as bind-opts}]
  (cond (= type "fanout")
        true

        (= type "direct")
        (throw (Exception. "DIRECT ROUTING NOT IMPLEMENTED."))

        (= type "topic")
        (cond (and (empty? routing-key)
                   (empty? key))
              true

              :else
              (match-pattern routing-key key))

        (= type "header")
        (every? (fn [[k pattern]]
                  (match-pattern pattern (get headers k)))
                (seq arguments))))

(defn add-in-atom
  "helper for adding queues and exchanges"
  {:added "3.0"}
  [atom keyword name opts defaults]
  (doto atom
    (swap! update-in [keyword] assoc name
           (clojure.core/atom {:meta (merge defaults opts)}))))

(defn add-exchange-atom
  "adds a queue to the atom
 
   (-> (atom {})
       (add-exchange-atom \"ex1\" {})
       (add-exchange-atom \"ex2\" {}) 
       ((comp sort keys :exchanges deref)))
   => [\"ex1\" \"ex2\"]"
  {:added "3.0"}
  [atom name opts]
  (add-in-atom atom :exchanges name opts common/*default-exchange-options*))

(defn add-queue-atom
  "adds a queue to the atom
 
   (-> (atom {})
       (add-queue-atom \"q3\" {})
       (add-queue-atom \"q2\" {}) 
       ((comp sort keys :queues deref)))
   => [\"q2\" \"q3\"]"
  {:added "3.0"}
  [atom name opts]
  (add-in-atom atom :queues name opts common/*default-queue-options*))

(defn list-in-atom
  "returns current list of queues
   (-> (atom {})
       (add-queue-atom \"q3\" {}) 
       (list-in-atom :queues))
   => {\"q3\" {:exclusive false
             :auto-delete false
             :durable false}}"
  {:added "3.0"}
  [atom keyword]
  (->> (keyword @atom)
       (reduce-kv (fn [out k atom]
                    (assoc out k (:meta @atom)))
                  {})))

(defn delete-in-atom
  "returns current list of queues
   (-> (atom {})
       (add-exchange-atom \"ex1\" {})
       (add-exchange-atom \"ex2\" {})
       (delete-in-atom :exchanges \"ex1\")
       (list-in-atom :exchanges))
   => {\"ex2\" {:type \"topic\"
              :internal false
              :auto-delete false
             :durable true}}"
  {:added "3.0"}
  [atom keyword name]
  (doto atom
    (swap! update-in [keyword] dissoc name)))

(defn list-bindings-atom
  "returns current list of exchanges
 
   (-> (atom {})
       (common/install-routing routes)
       (list-bindings-atom))
   => (contains-in {\"ex1\" {:exchanges {\"ex2\" [map?]}
                           :queues {\"q1\" [map?]}}
                    \"ex2\" {:queues {\"q2\" [map?]}}})"
  {:added "3.0"}
  [atom]
  (->> (:exchanges @atom)
       (reduce-kv (fn [out k atom]
                    (let [watches (.getWatches atom)]
                      (reduce-kv (fn [out bind watch]
                                   (let [{:keys [type dest] :as opt} (meta watch)
                                         type (keyword (str (name type) "s"))
                                         nopt (dissoc opt :type :source :dest)]
                                     (update-in out [k type dest] (fnil #(conj % nopt) []))))
                                 out
                                 watches)))
                  {})))

(defn bind-in-atom
  "helper function for binding to queues and exchanges"
  {:added "3.0"}
  [atom [source-key dest-key bind-type] source dest opts]
  (let [source-atom (get-in @atom [source-key source])
        dest-atom   (get-in @atom [dest-key dest])
        id          (or (:id opts) (str (java.util.UUID/randomUUID)))
        bind-opts   (merge common/*default-binding-options*
                           opts
                           {:id id :type bind-type :source source :dest dest})]
    (add-watch source-atom id
               (with-meta
                 (fn [_ _ _ {:keys [meta package] :as data}]
                   (if (route? meta package bind-opts)
                     (swap! dest-atom assoc-in [:package] package)))
                 bind-opts))
    atom))

(defn bind-exchange-atom
  "binds a queue to the exchange
 
   (-> (atom {})
       (common/install-routing routes)
       (add-exchange-atom \"ex3\" {})
       (bind-exchange-atom \"ex1\" \"ex3\" {})
       (list-bindings-atom))
   => (contains-in {\"ex1\" {:exchanges {\"ex2\" [map?]
                                       \"ex3\" [map?]}
                           :queues {\"q1\" [map?]}}
                   \"ex2\" {:queues {\"q2\" [map?]}}})"
  {:added "3.0"}
  [atom source dest opts]
  (bind-in-atom atom [:exchanges :exchanges :exchange] source dest opts))

(defn bind-queue-atom
  "binds an exchange to a queue
 
   (-> (atom {})
       (common/install-routing routes)
       (add-queue-atom \"q3\" {})
       (bind-queue-atom \"ex1\" \"q3\" {})
       (list-bindings-atom))
   => (contains-in {\"ex1\" {:exchanges {\"ex2\" [map?]}
                           :queues {\"q1\" [map?]
                                    \"q3\" [map?]}}
                   \"ex2\" {:queues {\"q2\" [map?]}}})"
  {:added "3.0"}
  [atom source dest opts]
  (bind-in-atom atom [:exchanges :queues :queue] source dest opts))

(defn list-consumers-atom
  "lists all connected consumers
 
   (-> (atom {})
       (common/install-routing routes)
       (add-consumer-atom  \"q2\" {:id :bar
                                 :sync true
                                 :function prn})
       (list-consumers-atom))
   => (contains-in {\"q1\" {}
                    \"q2\" {:bar {:id :bar,
                               :sync true,
                                :function fn?}}})"
  {:added "3.0"}
  [atom]
  (->> (:queues @atom)
       (reduce-kv (fn [out k v]
                    (assoc out k (->> (.getWatches v)
                                      (reduce-kv (fn [out k v]
                                                   (assoc out k (meta v)))
                                                 {}))))
                  {})))

(defn add-consumer-atom
  "adds a consumer to the queue"
  {:added "3.0"}
  [atom name {:keys [id function sync] :as handler}]
  (-> (get-in @atom [:queues name])
      (add-watch id (with-meta (fn [_ _ _ {:keys [package]}]
                                 (if sync
                                   (function (:body package))
                                   (future (function (:body package)))))
                      handler)))
  atom)

(defn delete-consumer-atom
  "deletes a consumer to the queue"
  {:added "3.0"}
  [atom name id]
  (-> (get-in @atom [:queues name])
      (remove-watch id))
  atom)

(defn publish-atom
  "publishes a message to the exchange
 
   (def p (promise))
   
   (-> (atom {})
       (common/install-routing routes)
       (add-consumer-atom \"q1\" {:id :bar
                                :sync true
                                :function #(deliver p %)})
       (publish-atom \"ex1\" \"hello there\" {}))
   
   @p => \"hello there\""
  {:added "3.0"}
  [atom exchange body {:keys [key headers] :as opts}]
  (let [exchange-atom (get-in @atom [:exchanges exchange])]
    (swap! exchange-atom assoc :package {:key key :headers headers :body body})
    atom))

(extend-protocol protocol.exchange/IExchange
  clojure.lang.Atom
  (-list-queues      [atom] (list-in-atom atom :queues))
  (-add-queue        [atom name opts] (add-queue-atom atom name opts))
  (-delete-queue     [atom name] (delete-in-atom atom :queues name))
  (-list-exchanges   [atom] (list-in-atom atom :exchanges))
  (-add-exchange     [atom name opts] (add-exchange-atom atom name opts))
  (-delete-exchange  [atom name] (delete-in-atom atom :exchanges name))
  (-list-bindings    [atom] (list-bindings-atom atom))
  (-bind-exchange    [atom source dest opts] (bind-exchange-atom atom source dest opts))
  (-bind-queue       [atom source dest opts] (bind-queue-atom atom source dest opts))
  (-list-consumers   [atom] (list-consumers-atom atom))
  (-add-consumer     [atom name handler] (add-consumer-atom atom name handler))
  (-delete-consumer  [atom name id] (delete-consumer-atom atom name id))
  (-publish          [atom exchange body opts] (publish-atom atom exchange body opts)))

(defrecord MockExchange [state]

  Object
  (toString [mq] (str "#exchange.mock"
                      (if state
                        (common/routing state {:short {}})
                        "<uninitiased>")))

  protocol.exchange/IExchange
  (-list-queues      [mq] (protocol.exchange/-list-queues state))
  (-add-queue        [mq name opts] (protocol.exchange/-add-queue state name opts) mq)
  (-delete-queue     [mq name] (protocol.exchange/-delete-queue state name) mq)
  (-list-exchanges   [mq] (protocol.exchange/-list-exchanges state))
  (-add-exchange     [mq name opts] (protocol.exchange/-add-exchange state name opts) mq)
  (-delete-exchange  [mq name] (protocol.exchange/-delete-exchange state name) mq)
  (-list-bindings    [mq] (protocol.exchange/-list-bindings state))
  (-bind-exchange    [mq source dest opts] (protocol.exchange/-bind-exchange state source dest opts) mq)
  (-bind-queue       [mq source dest opts] (protocol.exchange/-bind-queue state source dest opts) mq)
  (-list-consumers   [mq] (protocol.exchange/-list-consumers state))
  (-add-consumer     [mq name handler] (protocol.exchange/-add-consumer state name handler) mq)
  (-delete-consumer  [mq name id] (protocol.exchange/-delete-consumer state name id) mq)
  (-publish          [mq exchange body opts] (protocol.exchange/-publish state exchange body opts) mq)

  protocol.component/IComponent
  (-start [{:keys [routing consumers refresh file log] :as mq}]
    (cond-> mq
      file      (assoc-in [:file :transform] (comp common/routing atom))
      log       (assoc-in [:log :transform] (comp common/routing atom))
      :then     (table/attach-state)
      refresh   (common/purge-routing)
      routing   (common/install-routing routing)
      consumers (common/install-consumers consumers)))

  (-stop [mq]
    (table/detach-state mq)))

(defmethod print-method MockExchange
  [v w]
  (.write w (str v)))

(defmethod protocol.exchange/-create :mock
  [m]
  (map->MockExchange m))

==> platform-3.0.12.jar: hara/platform/exchange/base.clj <==
(ns hara.platform.exchange.base
  (:require [hara.protocol.exchange :as protocol.exchange]
            [hara.protocol.component :as protocol.component]
            [hara.core.component :as component]
            [hara.print.table :as table]
            [hara.platform.exchange.common :as common]))

(defn match-pattern
  "creates a re-pattern for the rabbitmq regex string
 
   (match-pattern \"*\" \"hello\")
   => true
 
   (match-pattern \".*.\" \".hello.\")
   => true"
  {:added "3.0"}
  [^String pattern input]
  (-> pattern
      (.replaceAll "\\." "\\\\\\.")
      (.replaceAll "\\*" ".*")
      (re-pattern)
      (re-find input)
      (= input)))

(defn route?
  "checks if a message will be routed
 
   (route? {:type \"fanout\"} {} {})
   => true
 
   (route? {:type \"topic\"}
           {:key \"user.account.login\"}
           {:routing-key \"*.account.*\"})
   => true
 
   (route? {:type \"header\"}
           {:headers {\"account\" \"login\"}}
           {:arguments {\"account\" \"*\"}})
   => true"
  {:added "3.0"}
  [{:keys [type] :as meta}
   {:keys [key headers] :as package}
   {:keys [routing-key arguments] :as bind-opts}]
  (cond (= type "fanout")
        true

        (= type "direct")
        (throw (Exception. "DIRECT ROUTING NOT IMPLEMENTED."))

        (= type "topic")
        (cond (and (empty? routing-key)
                   (empty? key))
              true

              :else
              (match-pattern routing-key key))

        (= type "header")
        (every? (fn [[k pattern]]
                  (match-pattern pattern (get headers k)))
                (seq arguments))))

(defn add-in-atom
  "helper for adding queues and exchanges"
  {:added "3.0"}
  [atom keyword name opts defaults]
  (doto atom
    (swap! update-in [keyword] assoc name
           (clojure.core/atom {:meta (merge defaults opts)}))))

(defn add-exchange-atom
  "adds a queue to the atom
 
   (-> (atom {})
       (add-exchange-atom \"ex1\" {})
       (add-exchange-atom \"ex2\" {}) 
       ((comp sort keys :exchanges deref)))
   => [\"ex1\" \"ex2\"]"
  {:added "3.0"}
  [atom name opts]
  (add-in-atom atom :exchanges name opts common/*default-exchange-options*))

(defn add-queue-atom
  "adds a queue to the atom
 
   (-> (atom {})
       (add-queue-atom \"q3\" {})
       (add-queue-atom \"q2\" {}) 
       ((comp sort keys :queues deref)))
   => [\"q2\" \"q3\"]"
  {:added "3.0"}
  [atom name opts]
  (add-in-atom atom :queues name opts common/*default-queue-options*))

(defn list-in-atom
  "returns current list of queues
   (-> (atom {})
       (add-queue-atom \"q3\" {}) 
       (list-in-atom :queues))
   => {\"q3\" {:exclusive false
             :auto-delete false
             :durable true}}"
  {:added "3.0"}
  [atom keyword]
  (->> (keyword @atom)
       (reduce-kv (fn [out k atom]
                    (assoc out k (:meta @atom)))
                  {})))

(defn delete-in-atom
  "returns current list of queues
   (-> (atom {})
       (add-exchange-atom \"ex1\" {})
       (add-exchange-atom \"ex2\" {})
       (delete-in-atom :exchanges \"ex1\")
       (list-in-atom :exchanges))
   => {\"ex2\" {:type \"topic\"
              :internal false
              :auto-delete false
             :durable true}}"
  {:added "3.0"}
  [atom keyword name]
  (doto atom
    (swap! update-in [keyword] dissoc name)))

(defn list-bindings-atom
  "returns current list of exchanges
 
   (-> (atom {})
       (common/install-routing routes)
       (list-bindings-atom))
   => (contains-in {\"ex1\" {:exchanges {\"ex2\" [map?]}
                           :queues {\"q1\" [map?]}}
                    \"ex2\" {:queues {\"q2\" [map?]}}})"
  {:added "3.0"}
  [atom]
  (->> (:exchanges @atom)
       (reduce-kv (fn [out k atom]
                    (let [watches (.getWatches ^clojure.lang.Atom atom)]
                      (reduce-kv (fn [out bind watch]
                                   (let [{:keys [type dest] :as opt} (meta watch)
                                         type (keyword (str (name type) "s"))
                                         nopt (dissoc opt :type :source :dest)]
                                     (update-in out [k type dest] (fnil #(conj % nopt) []))))
                                 out
                                 watches)))
                  {})))

(defn bind-in-atom
  "helper function for binding to queues and exchanges"
  {:added "3.0"}
  [atom [source-key dest-key bind-type] source dest opts]
  (let [source-atom (get-in @atom [source-key source])
        dest-atom   (get-in @atom [dest-key dest])
        id          (or (:id opts) (str (java.util.UUID/randomUUID)))
        bind-opts   (merge common/*default-binding-options*
                           opts
                           {:id id :type bind-type :source source :dest dest})]
    (add-watch source-atom id
               (with-meta
                 (fn [_ _ _ {:keys [meta package] :as data}]
                   (if (route? meta package bind-opts)
                     (swap! dest-atom assoc-in [:package] package)))
                 bind-opts))
    atom))

(defn bind-exchange-atom
  "binds a queue to the exchange
 
   (-> (atom {})
       (common/install-routing routes)
       (add-exchange-atom \"ex3\" {})
       (bind-exchange-atom \"ex1\" \"ex3\" {})
       (list-bindings-atom))
   => (contains-in {\"ex1\" {:exchanges {\"ex2\" [map?]
                                       \"ex3\" [map?]}
                           :queues {\"q1\" [map?]}}
                   \"ex2\" {:queues {\"q2\" [map?]}}})"
  {:added "3.0"}
  [atom source dest opts]
  (bind-in-atom atom [:exchanges :exchanges :exchange] source dest opts))

(defn bind-queue-atom
  "binds an exchange to a queue
 
   (-> (atom {})
       (common/install-routing routes)
       (add-queue-atom \"q3\" {})
       (bind-queue-atom \"ex1\" \"q3\" {})
       (list-bindings-atom))
   => (contains-in {\"ex1\" {:exchanges {\"ex2\" [map?]}
                           :queues {\"q1\" [map?]
                                    \"q3\" [map?]}}
                   \"ex2\" {:queues {\"q2\" [map?]}}})"
  {:added "3.0"}
  [atom source dest opts]
  (bind-in-atom atom [:exchanges :queues :queue] source dest opts))

(defn list-consumers-atom
  "lists all connected consumers
 
   (-> (atom {})
       (common/install-routing routes)
       (add-consumer-atom  \"q2\" {:id :bar
                                 :sync true
                                 :function prn})
       (list-consumers-atom))
   => (contains-in {\"q1\" {}
                    \"q2\" {:bar {:id :bar,
                               :sync true,
                                :function fn?}}})"
  {:added "3.0"}
  [atom]
  (->> (:queues @atom)
       (reduce-kv (fn [out k v]
                    (assoc out k (->> (.getWatches ^clojure.lang.Atom v)
                                      (reduce-kv (fn [out k v]
                                                   (assoc out k (meta v)))
                                                 {}))))
                  {})))

(defn add-consumer-atom
  "adds a consumer to the queue"
  {:added "3.0"}
  [atom name {:keys [id function sync] :as handler}]
  (-> (get-in @atom [:queues name])
      (add-watch id (with-meta (fn [_ _ _ {:keys [package]}]
                                 (if sync
                                   (function (:body package) (:headers package) {:key (:key package)})
                                   (future (function (:body package) (:headers package) {:key (:key package)}))))
                      handler)))
  atom)

(defn delete-consumer-atom
  "deletes a consumer to the queue"
  {:added "3.0"}
  [atom name id]
  (-> (get-in @atom [:queues name])
      (remove-watch id))
  atom)

(defn publish-atom
  "publishes a message to the exchange
 
   (def p (promise))
   
   (-> (atom {})
       (common/install-routing routes)
       (add-consumer-atom \"q1\" {:id :bar
                                :sync true
                                :function #(deliver p %)})
       (publish-atom \"ex1\" \"hello there\" {}))
   
   @p => \"hello there\""
  {:added "3.0"}
  ([atom exchange body {:keys [key headers] :as opts}]
   (if-let [exchange-atom (get-in @atom [:exchanges exchange])]
     (swap! exchange-atom assoc :package {:key key :headers headers :body body})
     (println "Unable to publish" {:exchange exchange :body body}))
   atom))

(extend-protocol protocol.exchange/IExchange
  clojure.lang.Atom
  (-list-queues      [atom] (list-in-atom atom :queues))
  (-add-queue        [atom name opts] (add-queue-atom atom name opts))
  (-delete-queue     [atom name] (delete-in-atom atom :queues name))
  (-list-exchanges   [atom] (list-in-atom atom :exchanges))
  (-add-exchange     [atom name opts] (add-exchange-atom atom name opts))
  (-delete-exchange  [atom name] (delete-in-atom atom :exchanges name))
  (-list-bindings    [atom] (list-bindings-atom atom))
  (-bind-exchange    [atom source dest opts] (bind-exchange-atom atom source dest opts))
  (-bind-queue       [atom source dest opts] (bind-queue-atom atom source dest opts))
  (-list-consumers   [atom] (list-consumers-atom atom))
  (-add-consumer     [atom name handler] (add-consumer-atom atom name handler))
  (-delete-consumer  [atom name id] (delete-consumer-atom atom name id))
  (-list-channels    [atom] [])
  (-add-channel      [atom] :not-supported)
  (-delete-channel   [atom id] :not-supported)
  (-publish          [atom exchange body opts] (publish-atom atom exchange body opts)))

(defrecord MockExchange [state]

  Object
  (toString [mq] (str "#exchange.mock"
                      (if state
                        (common/routing state {:short {}})
                        "<uninitiased>")))

  protocol.exchange/IExchange
  (-list-queues      [mq] (protocol.exchange/-list-queues state))
  (-add-queue        [mq name opts] (protocol.exchange/-add-queue state name opts) mq)
  (-delete-queue     [mq name] (protocol.exchange/-delete-queue state name) mq)
  (-list-exchanges   [mq] (protocol.exchange/-list-exchanges state))
  (-add-exchange     [mq name opts] (protocol.exchange/-add-exchange state name opts) mq)
  (-delete-exchange  [mq name] (protocol.exchange/-delete-exchange state name) mq)
  (-list-bindings    [mq] (protocol.exchange/-list-bindings state))
  (-bind-exchange    [mq source dest opts] (protocol.exchange/-bind-exchange state source dest opts) mq)
  (-bind-queue       [mq source dest opts] (protocol.exchange/-bind-queue state source dest opts) mq)
  (-list-consumers   [mq] (protocol.exchange/-list-consumers state))
  (-add-consumer     [mq name handler] (protocol.exchange/-add-consumer state name handler) mq)
  (-delete-consumer  [mq name id] (protocol.exchange/-delete-consumer state name id) mq)
  (-list-channels    [mq] (protocol.exchange/-list-channels state))
  (-add-channel      [mq] (protocol.exchange/-add-channel state))
  (-delete-channel   [mq id] (protocol.exchange/-delete-channel state id))
  (-publish          [mq exchange body opts] (protocol.exchange/-publish state exchange body opts) mq)

  protocol.component/IComponent
  (-start [{:keys [routing consumers refresh file log] :as mq}]
    (cond-> mq
      file      (assoc-in [:file :transform] (comp common/routing atom))
      log       (assoc-in [:log :transform] (comp common/routing atom))
      :then     (-> (table/attach-state)
                    (common/setup-routing)
                    (common/install-consumers))))
  
  (-stop [mq]
    (-> mq
        (common/remove-consumers)
        (common/teardown-routing)
        (table/detach-state))))

(defmethod print-method MockExchange
  [v ^java.io.Writer w]
  (.write w (str v)))

(defmethod protocol.exchange/-create :mock
  [m]
  (map->MockExchange m))
